import url from "url";
import * as uuid from "uuid";
import { EventEmitter } from "events";
import { WebSocket, WebSocketServer } from "ws";
import { ErrorCode, ErrorDescription, MessageTypeId } from "./constants.mjs";

/**
 * @typedef {Object} MessageCALL
 * @property {string} chargeBoxId - Charge Box Identifier
 * @property {2} messageTypeId - MessageTypeId
 * @property {string} uniqueId - MessageUniqueId
 * @property {Object} payload - Payload
 * @property {string} action - Action
 */

/**
 * @typedef {Object} MessageCALLRESULT
 * @property {string} chargeBoxId - Charge Box Identifier
 * @property {3} messageTypeId - MessageTypeId
 * @property {string} uniqueId - MessageUniqueId
 * @property {Object} payload - Payload
 */
/**
 * @typedef {Object} MessageCALLERROR
 * @property {string} chargeBoxId - Charge Box Identifier
 * @property {4} messageTypeId - MessageTypeId
 * @property {string} uniqueId - MessageUniqueId
 * @property {string} errorCode - Error Code
 * @property {string} errorDescription - Error Description
 * @property {Object} errorDetails - Error Detail
 */

/**
 * @callback ReplyCALL
 * @param {?CALLERROR} err CALLERROR
 * @param {{[name:string]: any}=} payload CALLRESULT Payload
 */

/**
 * @callback CALL_Listener
 * @param {MessageCALL} msg
 * @param {ReplyCALL} reply
 */
export class OCPP extends EventEmitter {
  /** @type {WebSocketServer} */
  wss;

  options;

  eventStacks = {};
  constructor(websocketServer, options = {}) {
    super();
    this.wss = websocketServer;
    if (!options.prefixUrl) {
      options.prefixUrl = "/";
    }
    this.options = options;
    this.init();
  }

  /**
   * @param {string} event
   * @param {CALL_Listener} listener
   */
  handle(event, listener) {
    super.on(`chargepoint:call:${event}`, listener);
  }

  init() {
    this.wss.on("connection", (socket, request) => {
      const cbid = this.pathToCBID(url.parse(request.url).pathname);
      socket.cbid = cbid;

      // set listener on CALL from Central System -> Charge Point
      this.on(`call:${cbid}`, (uniqueId, action, payload, resolve, reject) => {
        this.once(`callresult:${cbid}:${uniqueId}`, (...args) =>
          resolve(...args)
        );
        this.once(`callerror:${cbid}:${uniqueId}`, (...args) =>
          reject(...args)
        );

        socket.send(this.encodeCALL(uniqueId, action, payload));
      });

      this.emit(`connection`, cbid);

      socket.on("message", (buffer) => this.onMessageHandle(buffer, socket));
      socket.on("close", (code, reason) =>
        this.emit("close", cbid, code, reason)
      );
    });
  }

  onMessageHandle(buffer, socket) {
    let msg = this.serializeMessage(buffer.toString());
    msg.chargeBoxId = socket.cbid;
    switch (msg.messageTypeId) {
      case MessageTypeId.CALL:
        this.onChargePointCallHandler(msg, socket);
        break;
      case MessageTypeId.CALLRESULT:
        this.onChargePointCallResultHandler(msg);
        break;
      case MessageTypeId.CALLERROR:
        this.onChargePointCallErrorHandler(msg);
        break;
    }
  }

  /**
   * @param {MessageCALL} msg
   * @param {WebSocket} socket
   */
  onChargePointCallHandler(msg, socket) {
    const reply = (err, payload) => {
      if (err) {
        console.log(err);
        socket.send(
          this.encodeCALLERROR(msg.uniqueId, err.code, err.message, {})
        );
      } else {
        socket.send(this.encodeCALLRESULT(msg.uniqueId, payload));
      }
    };
    if (this.listenerCount(`chargepoint:call:${msg.action}`) > 0) {
      this.emit(`chargepoint:call:${msg.action}`, msg, reply);
    } else {
      reply(new ErrorNotImplement());
    }
  }

  /**
   * @param {MessageCALL} msg
   */
  onChargePointCallResultHandler(msg) {
    this.emit(`callresult:${msg.chargeBoxId}:${msg.uniqueId}`, msg);
    this.removeAllListeners(`callerror:${msg.chargeBoxId}:${msg.uniqueId}`);
  }

  /**
   * @param {MessageCALL} msg
   */
  onChargePointCallErrorHandler(msg) {
    this.emit(`callerror:${msg.chargeBoxId}:${msg.uniqueId}`, msg);
    this.removeAllListeners(`callresult:${msg.chargeBoxId}:${msg.uniqueId}`);
  }

  call(chargeBoxId, action, payload, uniqueId = uuid.v4()) {
    return new Promise((resolve, reject) => {
      try {
        this.emit(
          `call:${chargeBoxId}`,
          uniqueId,
          action,
          payload,
          resolve,
          reject
        );
      } catch (e) {
        reject(e);
      }
    });
  }

  encodeCALL(uniqueId, action, payload) {
    return JSON.stringify([MessageTypeId.CALL, uniqueId, action, payload]);
  }

  encodeCALLRESULT(uniqueId, payload) {
    return JSON.stringify([MessageTypeId.CALLRESULT, uniqueId, payload]);
  }

  encodeCALLERROR(
    uniqueId,
    errorCode,
    errorDescription = "",
    errorDetails = {}
  ) {
    return JSON.stringify([
      MessageTypeId.CALLERROR,
      uniqueId,
      errorCode,
      errorDescription,
      errorDetails,
    ]);
  }

  pathToCBID(path) {
    return path
      .replace(new RegExp(`^${this.options.prefixUrl}`, "i"), "")
      .trim();
  }

  /**
   * @param {*} jsonRaw
   * @returns {MessageCALL|MessageCALLRESULT|MessageCALLERROR=}
   */
  serializeMessage(jsonRaw) {
    const data = JSON.parse(jsonRaw);
    if (data.length >= 3 && data.length <= 5) {
      let msg = {
        messageTypeId: data[0],
        uniqueId: data[1],
      };
      switch (msg.messageTypeId) {
        case MessageTypeId.CALL:
          msg.action = data[2];
          msg.payload = data[3];
          break;
        case MessageTypeId.CALLRESULT:
          msg.payload = data[2];
          break;
        case MessageTypeId.CALLERROR:
          msg.errorCode = data[2];
          msg.errorDescription = data[3];
          msg.errorDetails = data[4];
          break;
      }
      return msg;
    } else {
      return null;
    }
  }
}

export class CALLERROR extends Error {
  constructor(errorCode) {
    let code = ErrorCode[errorCode];
    let message = ErrorDescription[code];
    if (!code) {
      code = ErrorCode.GenericError;
      message = errorCode;
    }
    super(message);
    this.code = code;
  }
}

export class ErrorNotImplement extends CALLERROR {
  constructor() {
    super(ErrorCode.NotImplemented);
  }
}
